/////////////////////////////////////////////////////////////////////
// PictureExWnd.h : Declaration of the CPictureExWnd
//
// Picture displaying control with support for the following formats:
// GIF (including animated GIF87a and GIF89a), JPEG, BMP, WMF, ICO, CUR
// 
// Written by Oleg Bykov (oleg_bykoff@rsdn.ru)
// Copyright (c) 2001
//
// To use CPictureExWnd, follow these steps:
//   - place a static control on your dialog (either a text or a bitmap)
//   - change its identifier to something else (e.g. IDC_MYPIC)
//   - add a member variable of type CPictureExWnd in your dialog class
//     (don't forget to #include "PictureEx.h" and add 
//     PictureEx.h and PictureEx.cpp to your project)
//   - in OnInitDialog() call the variable's SubclassWindow functions
//     (like this:
//         HWND hWnd = GetDlgItem(IDC_MYPIC);
//         if (hWnd) m_wndBanner.SubclassWindow(hWnd);
//     )
//   - call one of the overloaded CPictureEx::Load() functions somewhere
//     (OnInitDialog is a good place to do so)
//   - if the preceding Load() succeeded call Draw()
//  
// If your parent window is not a dialog, just add a member variable of type 
// CPictureExWnd to your main window class, call CPictureExWnd::Create()
// (say, in WM_CREATE handler function), then CPictureExWnd::Load() and 
// CPictureExWnd::Draw().
//
// By default, the control initializes its background to COLOR_3DFACE
// (see CPictureEx::PrepareDC()). You can change the background by
// calling CPictureEx::SetBkColor(COLORREF) after CPictureEx::Load().
//
// I decided to leave in the class the functions to write separate frames from 
// animated GIF to disk. If you want to use them, uncomment #define GIF_TRACING 
// and an appropriate section in CPictureEx::Load(HGLOBAL, DWORD). These functions 
// won't be compiled and linked to your project unless you uncomment #define GIF_TRACING,
// so you don't have to worry.
// 
// Warning: this code hasn't been subject to a heavy testing, so
// use it on your own risk. The author accepts no liability for the 
// possible damage caused by this code.
//
//
// Version 1.0  7 Aug 2001
//              Initial release
//
// Version 1.1  6 Sept 2001
//              ATL version of the class
//
// Version 1.2  14 Oct 2001
//              - Fixed a problem with loading GIFs from resources
//                in MFC-version of the class for multi-modules apps.
//                Thanks to Ruben Avila-Carretero for finding this out.
//
//              - Got rid of waitable timer in ThreadAnimation()
//                Now CPictureEx[Wnd] works in Win95 too.
//                Thanks to Alex Egiazarov and Wayne King for the idea.
//
//              - Fixed a visual glitch of using SetBkColor.
//                Thanks to Kwangjin Lee for finding this out.
//
// Version 1.3  10 Nov 2001
//              - Fixed a DC leak. One DC leaked per each UnLoad()
//                (forgot to put a ReleaseDC() in the end of 
//                CPictureExWnd::PrepareDC() function).
//
//              - Now it is possible to set a clipping rectangle using
//                CPictureEx[Wnd]::SetPaintRect(const LPRECT) function.
//                The LPRECT parameter tells the class what portion of
//                a picture should it display. If the clipping rect is 
//                not set, the whole picture is shown.
//                Thanks to Fabrice Rodriguez for the idea.
//
//              - Added support for Stop/Draw. Now you can Stop() an
//                animated GIF, then Draw() it again, it will continue
//                animation from the frame it was stopped on. You can 
//                also know if a GIF is currently playing with the 
//                IsPlaying() function.
//             
//              - Got rid of math.h and made m_bExitThread volatile. 
//                Thanks to Piotr Sawicki for the suggestion.
//
/////////////////////////////////////////////////////////////////////

#ifndef __PICTUREEXWND_INC__
#define __PICTUREEXWND_INC__

//#include "olectl.inc"
//#include "commctrl.inc"
//#include <vector>
#include "include.inc"


//#define GIF_TRACING  // uncomment it if you want detailed TRACEs


struct TFrame    // structure that keeps a single frame info
{
	IPicture *m_pPicture;  // pointer to the interface used for drawing
	SIZE     m_frameSize;
	SIZE     m_frameOffset;
	UINT     m_nDelay;     // delay (in 1/100s of a second)
	UINT     m_nDisposal;  // disposal method
//sapero added
	TFrame  *next;
};

//#pragma pack(1)   // turn byte alignment on

#typedef GIFBlockTypes int
const BLOCK_UNKNOWN = 0;
const BLOCK_APPEXT = 1;
const BLOCK_COMMEXT = 2;
const BLOCK_CONTROLEXT = 3;
const BLOCK_PLAINTEXT = 4;
const BLOCK_IMAGE = 5;
const BLOCK_TRAILER = 6;


#typedef ControlExtValues int // graphic control extension packed field values
const GCX_PACKED_DISPOSAL = 0;  // disposal method
const GCX_PACKED_USERINPUT = 1;
const GCX_PACKED_TRANSPCOLOR = 2;


#typedef LSDPackedValues int // logical screen descriptor packed field values
const LSD_PACKED_GLOBALCT = 0;
const LSD_PACKED_CRESOLUTION = 1;
const LSD_PACKED_SORT = 2;
const LSD_PACKED_GLOBALCTSIZE = 3;


#typedef IDPackedValues int  // image descriptor packed field values
const ID_PACKED_LOCALCT = 0;
const ID_PACKED_INTERLACE = 1;
const ID_PACKED_SORT = 2;
const ID_PACKED_LOCALCTSIZE = 3;


struct TGIFHeader,1       // GIF header  
{
	dstring m_cSignature[3]; // Signature - Identifies the GIF Data Stream
						  // This field contains the fixed value 'GIF'
	dstring m_cVersion[3];	// Version number. May be one of the following:
						// "87a" or "89a"
};

struct TGIFLSDescriptor,1 // Logical Screen Descriptor
{
	unsigned word m_wWidth;	// 2 bytes. Logical screen width
	unsigned word m_wHeight; // 2 bytes. Logical screen height

	unsigned byte m_cPacked;      // packed field	

	unsigned byte m_cBkIndex;     // 1 byte. Background color index
	unsigned byte m_cPixelAspect; // 1 byte. Pixel aspect ratio
	//declare GetPackedValue(LSDPackedValues Value),int;
};

struct TGIFAppExtension,1 // application extension block
{
	unsigned byte m_cExtIntroducer; // extension introducer (0x21)
	unsigned byte m_cExtLabel; // app. extension label (0xFF)
	unsigned byte m_cBlockSize; // fixed value of 11
	dstring m_cAppIdentifier[8];   // application identifier
	dstring m_cAppAuth[3];  // application authentication code
};

struct TGIFControlExt,1 // graphic control extension block
{
	unsigned byte m_cExtIntroducer; // extension introducer (0x21)
	unsigned byte m_cControlLabel;  // control extension label (0xF9)
	unsigned byte m_cBlockSize; // fixed value of 4
	unsigned byte m_cPacked;    // packed field
	unsigned word m_wDelayTime;	// delay time
	unsigned byte m_cTColorIndex; // transparent color index
	unsigned byte m_cBlockTerm;   // block terminator (0x00)

	//declare GetPackedValue(ControlExtValues Value),int;
};

struct TGIFCommentExt,1  // comment extension block
{
	unsigned byte m_cExtIntroducer; // extension introducer (0x21)
	unsigned byte m_cCommentLabel;  // comment extension label (0xFE)
};

struct TGIFPlainTextExt,1 // plain text extension block
{
	unsigned byte m_cExtIntroducer;  // extension introducer (0x21)
	unsigned byte m_cPlainTextLabel; // text extension label (0x01)
	unsigned byte m_cBlockSize; // fixed value of 12
	unsigned word m_wLeftPos;    // text grid left position
	unsigned word m_wTopPos;     // text grid top position
	unsigned word m_wGridWidth;  // text grid width
	unsigned word m_wGridHeight; // text grid height
	unsigned byte m_cCellWidth;  // character cell width
	unsigned byte m_cCellHeight; // character cell height
	unsigned byte m_cFgColor; // text foreground color index
	unsigned byte m_cBkColor; // text background color index
};

struct TGIFImageDescriptor,1 // image descriptor block
{
	unsigned byte m_cImageSeparator; // image separator byte (0x2C)
	unsigned word m_wLeftPos; // image left position
	unsigned word m_wTopPos;  // image top position
	unsigned word m_wWidth;   // image width
	unsigned word m_wHeight;  // image height
	unsigned byte m_cPacked; // packed field
	//declare GetPackedValue(IDPackedValues Value),int;
};


/////////////////////////////////////////////////////////////////////////////
// CPictureExWnd
class CPictureExWnd : CWindow
{
	//DECLARE_WND_SUPERCLASS(TEXT("PICTUREEXWND"), TEXT("STATIC"))

	//BEGIN_MSG_MAP(CPictureExWnd)
	//	MESSAGE_HANDLER(WM_DESTROY, OnDestroy)
	//	MESSAGE_HANDLER(WM_PAINT, OnPaint)
	//END_MSG_MAP()

	declare CPictureExWnd();
	declare _CPictureExWnd();

	declare GetPaintRect(RECT *lpRect),BOOL;
	declare SetPaintRect(RECT *lpRect),BOOL;
	declare Stop();   // stops animation
	declare UnLoad(); // stops animation plus releases all resources

	declare IsGIF(),BOOL;
	declare IsPlaying(),BOOL;
	declare IsAnimatedGIF(),BOOL;
	declare GetSize(),SIZE;
	declare GetFrameCount(),int;
	declare GetBkColor(),COLORREF;
	declare SetBkColor(COLORREF clr);

	// draws the picture (starts an animation thread if needed)
	// if an animation was previously stopped by Stop(),
	// continues it from the last displayed frame
	declare Draw(),BOOL;

	// loads a picture from a file
	// i.e. Load(_T("mypic.gif"));
	declare LoadImage(LPCTSTR szFileName),BOOL;

	// loads a picture from a global memory block (allocated by GlobalAlloc)
	// Warning: this function DOES NOT free the global memory, pointed to by hGlobal
	declare Load(HGLOBAL hGlobal, DWORD dwSize),BOOL;

	// loads a picture from a program resource
	// i.e. Load(MAKEINTRESOURCE(IDR_MYPIC),_T("GIFTYPE"));
	declare LoadFromResource(LPCTSTR szResourceName,LPCTSTR szResourceType),BOOL;


#ifdef GIF_TRACING
	declare EnumGIFBlocks();
	declare WriteDataOnDisk(LPCTSTR szFileName, HGLOBAL hData, DWORD dwSize);
#endif

	RECT m_PaintRect;
	SIZE m_PictureSize;
	COLORREF m_clrBackground;
	UINT m_nDataSize;
	UINT m_nCurrFrame;
	UINT m_nCurrOffset;
	UINT m_nGlobalCTSize;
	BOOL m_bIsGIF;
	BOOL m_bIsPlaying;
	BOOL m_bIsInitialized;

	HDC m_hMemDC;

	HDC m_hDispMemDC;
	HBITMAP m_hDispMemBM;
	HBITMAP m_hDispOldBM;

	HBITMAP m_hBitmap2;
	HBITMAP m_hOldBitmap;
	IPicture * m_pPicture;
	TGIFHeader * m_pGIFHeader;
	unsigned byte * m_pRawData;
	TGIFLSDescriptor * m_pGIFLSDescriptor;
//sapero changed
//	std::vector<TFrame> m_arrFrames;
	TFrame *m_arrFrames;
	TFrame *m_arrLastFrame;
	int m_arrFrames_size;  // number of structures in list

	//BOOL m_bExitThread;
	//HANDLE m_hThread;
	//HANDLE m_hExitEvent;
	//declare ThreadAnimation();
	//declare _ThreadAnimation(),UINT; // changed static fn to method and removed argument

	TFrame *m_arrCurrentFrame;
	declare StartAnimation();
	declare OnBeforeTimer();
	declare virtual OnTimer(int nIDEvent);

	declare GetNextBlockLen(),int;
	declare SkipNextBlock(),BOOL;
	declare SkipNextGraphicBlock(),BOOL;
	declare PrepareDC(int nWidth, int nHeight),BOOL;
	declare ResetDataPointer();
	declare GetNextBlock(),GIFBlockTypes;
	declare GetSubBlocksLen(UINT nStartingOffset),UINT;
	declare GetNextGraphicBlock(UINT *pBlockLen, UINT *pDelay, SIZE *pBlockSize, SIZE *pBlockOffset, UINT *pDisposal),HGLOBAL;
	declare Invalidate();
	declare virtual OnDestroy();
	declare virtual OnPaint();
};

#endif